package ru.dbotthepony.kstarbound.defs.dungeon

import com.google.common.collect.ImmutableList
import ru.dbotthepony.kommons.arrays.Object2DArray
import ru.dbotthepony.kommons.math.RGBAColor
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.defs.image.Image

class ImagePartReader(part: DungeonPart, val images: ImmutableList<Image>) : PartReader(part) {
	override val size: Vector2i
		get() = if (images.isEmpty()) Vector2i.ZERO else images.first().size

	// it is much cheaper to just read all images and store 2D array
	// of references than loading / keeping images themselves around
	// `Image` class doesn't actually keep pixel data around for too long,
	// if it doesn't get accessed in some time it gets purged from ram
	private val layers = Array(images.size) {
		Object2DArray.nulls<DungeonTile>(images[it].width, images[it].height)
	} as Array<Object2DArray<DungeonTile>>

	override fun bind(def: DungeonDefinition) {
		check(def.tiles.isNotEmpty) { "Image parts require 'tiles' palette to be present in .dungeon definition" }

		for ((i, image) in images.withIndex()) {
			val layer = layers[i]

			for (y in 0 until image.height) {
				for (x in 0 until image.width) {
					val color = image[x, y]
					val tile = part.dungeon.tiles[color]

					if (tile == null) {
						val parse = RGBAColor.abgr(color)
						throw IllegalStateException("Unknown tile on ${image.path} at $x, $y: [${parse.redInt}, ${parse.greenInt}, ${parse.blueInt}, ${parse.alphaInt}] (index $color)")
					}

					layer[x, y] = tile
				}
			}
		}
	}

	override fun <T> walkTiles(callback: TileCallback<T>): KOptional<T> {
		for (layer in layers) {
			for (y in 0 until layer.rows) {
				for (x in 0 until layer.columns) {
					val get = callback(x, y, layer[x, y])
					if (get.isPresent) return get
				}
			}
		}

		return KOptional()
	}

	override fun <T> walkTilesAt(x: Int, y: Int, callback: TileCallback<T>): KOptional<T> {
		for (layer in layers) {
			if (x in 0 until layer.columns && y in 0 until layer.rows) {
				val get = callback(x, y, layer[x, y])
				if (get.isPresent) return get
			}
		}

		return KOptional()
	}
}
